@@ -5,12 +5,10 @@
 #include "qqmlcomponent_p.h"
 #include "qqmlcomponentattached_p.h"
 
-#include "qqmlcontext_p.h"
 #include "qqmlengine_p.h"
 #include "qqmlvme_p.h"
 #include "qqml.h"
 #include "qqmlengine.h"
-#include "qqmlbinding_p.h"
 #include "qqmlincubator.h"
 #include "qqmlincubator_p.h"
 #include <private/qqmljavascriptexpression_p.h>
@@ -303,19 +301,14 @@ void QQmlComponentPrivate::fromTypeData(const QQmlRefPointer<QQmlTypeData> &data
 
     if (!compilationUnit) {
         Q_ASSERT(data->isError());
-        state.errors = data->errors();
+        state.errors.clear();
+        state.appendErrors(data->errors());
     }
 }
 
-RequiredProperties &QQmlComponentPrivate::requiredProperties()
-{
-    Q_ASSERT(state.creator);
-    return state.creator->requiredProperties();
-}
-
 bool QQmlComponentPrivate::hadTopLevelRequiredProperties() const
 {
-    return state.creator->componentHadTopLevelRequiredProperties();
+    return state.creator()->componentHadTopLevelRequiredProperties();
 }
 
 void QQmlComponentPrivate::clear()
@@ -326,6 +319,8 @@ void QQmlComponentPrivate::clear()
     }
 
     compilationUnit.reset();
+    loadedType = {};
+    isInlineComponent = false;
 }
 
 QObject *QQmlComponentPrivate::doBeginCreate(QQmlComponent *q, QQmlContext *context)
@@ -359,15 +354,19 @@ bool QQmlComponentPrivate::setInitialProperty(
         segment = scope.engine->newString(properties.last());
         object->put(segment, scope.engine->metaTypeToJS(value.metaType(), value.constData()));
         if (scope.engine->hasException) {
-            state.errors.push_back(scope.engine->catchExceptionAsQmlError());
+            qmlWarning(base, scope.engine->catchExceptionAsQmlError());
             scope.engine->hasException = false;
             return false;
         }
         return true;
     }
 
-    QQmlProperty prop = QQmlComponentPrivate::removePropertyFromRequired(
-            base, name, requiredProperties(), engine);
+    QQmlProperty prop;
+    if (state.hasUnsetRequiredProperties())
+        prop = QQmlComponentPrivate::removePropertyFromRequired(
+                    base, name, state.requiredProperties(), engine);
+    else
+        prop = QQmlProperty(base, name, engine);
     QQmlPropertyPrivate *privProp = QQmlPropertyPrivate::get(prop);
     const bool isValid = prop.isValid();
     if (!isValid || !privProp->writeValueProperty(value, {})) {
@@ -380,7 +379,7 @@ bool QQmlComponentPrivate::setInitialProperty(
                                                 "%2 does not have a property called %1")
                                          .arg(name, QQmlMetaType::prettyTypeName(base)));
         }
-        state.errors.push_back(error);
+        qmlWarning(base, error);
         return false;
     }
 
@@ -403,17 +402,17 @@ QQmlComponent::~QQmlComponent()
 {
     Q_D(QQmlComponent);
 
-    if (d->state.completePending) {
+    if (d->state.isCompletePending()) {
         qWarning("QQmlComponent: Component destroyed while completion pending");
 
         if (isError()) {
             qWarning() << "This may have been caused by one of the following errors:";
-            for (const QQmlError &error : std::as_const(d->state.errors))
-                qWarning().nospace().noquote() << QLatin1String("    ") << error;
+            for (const QQmlComponentPrivate::AnnotatedQmlError &e : std::as_const(d->state.errors))
+                qWarning().nospace().noquote() << QLatin1String("    ") << e.error;
         }
 
         // we might not have the creator anymore if the engine is gone
-        if (d->state.creator)
+        if (d->state.hasCreator())
             d->completeCreate();
     }
 
@@ -450,7 +449,7 @@ QQmlComponent::Status QQmlComponent::status() const
         return Loading;
     else if (!d->state.errors.isEmpty())
         return Error;
-    else if (d->engine && d->compilationUnit)
+    else if (d->engine && (d->compilationUnit || d->loadedType.isValid()))
         return Ready;
     else
         return Null;
@@ -491,6 +490,8 @@ bool QQmlComponent::isLoading() const
 /*!
     Returns true if the component was created in a QML files that specifies
     \c{pragma ComponentBehavior: Bound}, otherwise returns false.
+
+    \since 6.5
  */
 bool QQmlComponent::isBound() const
 {
@@ -539,7 +540,7 @@ QQmlComponent::QQmlComponent(QQmlEngine *engine, QObject *parent)
     Q_D(QQmlComponent);
     d->engine = engine;
     QObject::connect(engine, &QObject::destroyed, this, [d]() {
-        d->state.creator.reset();
+        d->state.clear();
         d->engine = nullptr;
     });
 }
@@ -574,6 +575,36 @@ QQmlComponent::QQmlComponent(QQmlEngine *engine, const QUrl &url, CompilationMod
     d->loadUrl(url, mode);
 }
 
+/*!
+    Create a QQmlComponent from the given \a uri and \a typeName and give it
+    the specified \a parent and \a engine. If possible, the component will
+    be loaded synchronously.
+
+    \sa loadFromModule()
+    \since 6.5
+    \overload
+*/
+QQmlComponent::QQmlComponent(QQmlEngine *engine, QAnyStringView uri, QAnyStringView typeName, QObject *parent)
+    : QQmlComponent(engine, uri, typeName, QQmlComponent::PreferSynchronous, parent)
+{
+
+}
+
+/*!
+    Create a QQmlComponent from the given \a uri and \a typeName and give it
+    the specified \a parent and \a engine. If \a mode is \l Asynchronous,
+    the component will be loaded and compiled asynchronously.
+
+    \sa loadFromModule()
+    \since 6.5
+    \overload
+*/
+QQmlComponent::QQmlComponent(QQmlEngine *engine, QAnyStringView uri, QAnyStringView typeName, CompilationMode mode, QObject *parent)
+    : QQmlComponent(engine, parent)
+{
+    loadFromModule(uri, typeName, mode);
+}
+
 /*!
     Create a QQmlComponent from the given \a fileName and give it the specified
     \a parent and \a engine.
@@ -598,8 +629,12 @@ QQmlComponent::QQmlComponent(QQmlEngine *engine, const QString &fileName,
     : QQmlComponent(engine, parent)
 {
     Q_D(QQmlComponent);
-    const QUrl url = QDir::isAbsolutePath(fileName) ? QUrl::fromLocalFile(fileName) : QUrl(fileName);
-    d->loadUrl(url, mode);
+    if (fileName.startsWith(u':'))
+        d->loadUrl(QUrl(QLatin1String("qrc") + fileName), mode);
+    else if (QDir::isAbsolutePath(fileName))
+        d->loadUrl(QUrl::fromLocalFile(fileName), mode);
+    else
+        d->loadUrl(QUrl(fileName), mode);
 }
 
 /*!
@@ -720,7 +755,7 @@ void QQmlComponentPrivate::loadUrl(const QUrl &newUrl, QQmlComponent::Compilatio
     if (newUrl.isEmpty()) {
         QQmlError error;
         error.setDescription(QQmlComponent::tr("Invalid empty URL"));
-        state.errors << error;
+        state.errors.emplaceBack(error);
         return;
     }
 
@@ -732,7 +767,6 @@ void QQmlComponentPrivate::loadUrl(const QUrl &newUrl, QQmlComponent::Compilatio
     QQmlTypeLoader::Mode loaderMode = (mode == QQmlComponent::Asynchronous)
             ? QQmlTypeLoader::Asynchronous
             : QQmlTypeLoader::PreferSynchronous;
-
     QQmlRefPointer<QQmlTypeData> data = QQmlEnginePrivate::get(engine)->typeLoader.getType(url, loaderMode);
 
     if (data->isCompleteOrError()) {
@@ -756,10 +790,11 @@ void QQmlComponentPrivate::loadUrl(const QUrl &newUrl, QQmlComponent::Compilatio
 QList<QQmlError> QQmlComponent::errors() const
 {
     Q_D(const QQmlComponent);
-    if (isError())
-        return d->state.errors;
-    else
-        return QList<QQmlError>();
+    QList<QQmlError> errors;
+    errors.reserve(d->state.errors.size());
+    for (const QQmlComponentPrivate::AnnotatedQmlError &annotated : d->state.errors)
+        errors.emplaceBack(annotated.error);
+    return errors;
 }
 
 /*!
@@ -783,7 +818,8 @@ QString QQmlComponent::errorString() const
     QString ret;
     if(!isError())
         return ret;
-    for (const QQmlError &e : d->state.errors) {
+    for (const QQmlComponentPrivate::AnnotatedQmlError &annotated : d->state.errors) {
+        const QQmlError &e = annotated.error;
         ret += e.url().toString() + QLatin1Char(':') +
                QString::number(e.line()) + QLatin1Char(' ') +
                e.description() + QLatin1Char('\n');
@@ -834,23 +870,7 @@ QQmlComponent::QQmlComponent(QQmlComponentPrivate &dd, QObject *parent)
 QObject *QQmlComponent::create(QQmlContext *context)
 {
     Q_D(QQmlComponent);
-
-    QObject *rv = d->doBeginCreate(this, context);
-    if (rv) {
-        completeCreate();
-    } else if (d->state.completePending) {
-        // overridden completCreate might assume that
-        // the object has actually been created
-        ++creationDepth;
-        QQmlEnginePrivate *ep = QQmlEnginePrivate::get(d->engine);
-        d->complete(ep, &d->state);
-        --creationDepth;
-    }
-    if (rv && !d->requiredProperties().empty()) {
-        delete  rv;
-        return nullptr;
-    }
-    return rv;
+    return d->createWithProperties(nullptr, QVariantMap {}, context);
 }
 
 /*!
@@ -861,9 +881,9 @@ QObject *QQmlComponent::create(QQmlContext *context)
     TODO: also mention errorString() when QTBUG-93239 is fixed
     \endomit
 
-    If any of the \c initialProperties cannot be set, \l isError() will return
-    \c true, and the \l errors() function can be used to
-    get detailed information about the error(s).
+    If any of the \a initialProperties cannot be set, a warning is issued. If
+    there are unset required properties, the object creation fails and returns
+    \c nullptr, in which case \l isError() will return \c true.
 
     \sa QQmlComponent::create
     \since 5.14
@@ -871,16 +891,46 @@ QObject *QQmlComponent::create(QQmlContext *context)
 QObject *QQmlComponent::createWithInitialProperties(const QVariantMap& initialProperties, QQmlContext *context)
 {
     Q_D(QQmlComponent);
+    return d->createWithProperties(nullptr, initialProperties, context);
+}
 
-    QObject *rv = d->doBeginCreate(this, context);
-    if (rv) {
-        setInitialProperties(rv, initialProperties);
-        completeCreate();
-    }
-    if (!d->requiredProperties().empty()) {
-        d->requiredProperties().clear();
+static void QQmlComponent_setQmlParent(QObject *me, QObject *parent); // forward declaration
+
+/*! \internal
+ */
+QObject *QQmlComponentPrivate::createWithProperties(QObject *parent, const QVariantMap &properties,
+                                                    QQmlContext *context, CreateBehavior behavior)
+{
+    Q_Q(QQmlComponent);
+
+    QObject *rv = doBeginCreate(q, context);
+    if (!rv) {
+        if (state.isCompletePending()) {
+            // overridden completCreate might assume that
+            // the object has actually been created
+            ++creationDepth;
+            QQmlEnginePrivate *ep = QQmlEnginePrivate::get(engine);
+            complete(ep, &state);
+            --creationDepth;
+        }
         return nullptr;
     }
+
+    QQmlComponent_setQmlParent(rv, parent); // internally checks if parent is nullptr
+
+    q->setInitialProperties(rv, properties);
+    q->completeCreate();
+
+    if (state.hasUnsetRequiredProperties()) {
+        if (behavior == CreateWarnAboutRequiredProperties) {
+            for (const auto &unsetRequiredProperty : std::as_const(*state.requiredProperties())) {
+                const QQmlError error = unsetRequiredPropertyToQQmlError(unsetRequiredProperty);
+                qmlWarning(rv, error);
+            }
+        }
+        delete rv;
+        rv = nullptr;
+    }
     return rv;
 }
 
@@ -925,7 +975,7 @@ QObject *QQmlComponentPrivate::beginCreate(QQmlRefPointer<QQmlContextData> conte
     auto cleanup = qScopeGuard([this] {
         if (!state.errors.isEmpty() && lcQmlComponentGeneral().isDebugEnabled()) {
             for (const auto &e : std::as_const(state.errors)) {
-                qCDebug(lcQmlComponentGeneral) << "QQmlComponent: " << e.toString();
+                qCDebug(lcQmlComponentGeneral) << "QQmlComponent: " << e.error.toString();
             }
         }
     });
@@ -944,11 +994,20 @@ QObject *QQmlComponentPrivate::beginCreate(QQmlRefPointer<QQmlContextData> conte
         return nullptr;
     }
 
-    if (state.completePending) {
+    if (state.isCompletePending()) {
         qWarning("QQmlComponent: Cannot create new component instance before completing the previous");
         return nullptr;
     }
 
+    // filter out temporary errors as they do not really affect component's
+    // state (they are not part of the document compilation)
+    state.errors.erase(std::remove_if(state.errors.begin(), state.errors.end(),
+                                      [](const QQmlComponentPrivate::AnnotatedQmlError &e) {
+                                          return e.isTransient;
+                                      }),
+                       state.errors.end());
+    state.clearRequiredProperties();
+
     if (!q->isReady()) {
         qWarning("QQmlComponent: Component is not ready");
         return nullptr;
@@ -965,15 +1024,29 @@ QObject *QQmlComponentPrivate::beginCreate(QQmlRefPointer<QQmlContextData> conte
 
     enginePriv->inProgressCreations++;
     state.errors.clear();
-    state.completePending = true;
+    state.setCompletePending(true);
 
-    enginePriv->referenceScarceResources();
     QObject *rv = nullptr;
-    state.creator.reset(new QQmlObjectCreator(std::move(context), compilationUnit, creationContext));
-    rv = state.creator->create(start);
-    if (!rv)
-        state.errors = state.creator->errors;
-    enginePriv->dereferenceScarceResources();
+
+    if (!loadedType.isValid()) {
+        enginePriv->referenceScarceResources();
+        state.initCreator(std::move(context), compilationUnit, creationContext);
+        rv = state.creator()->create(start, nullptr, nullptr, isInlineComponent ? QQmlObjectCreator::InlineComponent : QQmlObjectCreator::NormalObject);
+        if (!rv)
+            state.appendCreatorErrors();
+        enginePriv->dereferenceScarceResources();
+    } else {
+        rv = loadedType.createWithQQmlData();
+        QQmlPropertyCache::ConstPtr propertyCache = QQmlData::ensurePropertyCache(rv);
+        for (int i = 0, propertyCount = propertyCache->propertyCount(); i < propertyCount; ++i) {
+            if (const QQmlPropertyData *propertyData = propertyCache->property(i); propertyData->isRequired()) {
+                state.ensureRequiredPropertyStorage();
+                RequiredPropertyInfo info;
+                info.propertyName = propertyData->name(rv);
+                state.addPendingRequiredProperty(rv, propertyData, info);
+            }
+        }
+    }
 
     if (rv) {
         QQmlData *ddata = QQmlData::get(rv);
@@ -1000,14 +1073,15 @@ void QQmlComponentPrivate::beginDeferred(QQmlEnginePrivate *enginePriv,
         enginePriv->inProgressCreations++;
 
         ConstructionState state;
-        state.completePending = true;
+        state.setCompletePending(true);
 
-        state.creator.reset(new QQmlObjectCreator(
-                                 deferredData->context->parent(), deferredData->compilationUnit,
-                                 QQmlRefPointer<QQmlContextData>()));
+        auto creator = state.initCreator(
+                    deferredData->context->parent(),
+                    deferredData->compilationUnit,
+                    QQmlRefPointer<QQmlContextData>());
 
-        if (!state.creator->populateDeferredProperties(object, deferredData))
-            state.errors << state.creator->errors;
+        if (!creator->populateDeferredProperties(object, deferredData))
+            state.appendCreatorErrors();
         deferredData->bindings.clear();
 
         deferredState->push_back(std::move(state));
@@ -1022,11 +1096,11 @@ void QQmlComponentPrivate::completeDeferred(QQmlEnginePrivate *enginePriv, QQmlC
 
 void QQmlComponentPrivate::complete(QQmlEnginePrivate *enginePriv, ConstructionState *state)
 {
-    if (state->completePending) {
+    if (state->isCompletePending()) {
         QQmlInstantiationInterrupt interrupt;
-        state->creator->finalize(interrupt);
+        state->creator()->finalize(interrupt);
 
-        state->completePending = false;
+        state->setCompletePending(false);
 
         enginePriv->inProgressCreations--;
 
@@ -1043,6 +1117,7 @@ void QQmlComponentPrivate::complete(QQmlEnginePrivate *enginePriv, ConstructionS
     Finds the matching top-level property with name \a name of the component \a createdComponent.
     If it was a required property or an alias to a required property contained in \a
     requiredProperties, it is removed from it.
+    \a requiredProperties must be non-null.
 
     If wasInRequiredProperties is non-null, the referenced boolean is set to true iff the property
     was found in requiredProperties.
@@ -1055,9 +1130,10 @@ void QQmlComponentPrivate::complete(QQmlEnginePrivate *enginePriv, ConstructionS
     setInitialProperties.
  */
 QQmlProperty QQmlComponentPrivate::removePropertyFromRequired(
-        QObject *createdComponent, const QString &name, RequiredProperties &requiredProperties,
+        QObject *createdComponent, const QString &name, RequiredProperties *requiredProperties,
         QQmlEngine *engine, bool *wasInRequiredProperties)
 {
+    Q_ASSERT(requiredProperties);
     QQmlProperty prop(createdComponent, name, engine);
     auto privProp = QQmlPropertyPrivate::get(prop);
     if (prop.isValid()) {
@@ -1078,11 +1154,11 @@ QQmlProperty QQmlComponentPrivate::removePropertyFromRequired(
             Q_ASSERT(data && data->propertyCache);
             targetProp = data->propertyCache->property(targetProp->coreIndex());
         }
-        auto it = requiredProperties.find(targetProp);
-        if (it != requiredProperties.end()) {
+        auto it = requiredProperties->find({createdComponent, targetProp});
+        if (it != requiredProperties->end()) {
             if (wasInRequiredProperties)
                 *wasInRequiredProperties = true;
-            requiredProperties.erase(it);
+            requiredProperties->erase(it);
         } else {
             if (wasInRequiredProperties)
                 *wasInRequiredProperties = false;
@@ -1110,12 +1186,21 @@ void QQmlComponent::completeCreate()
 
 void QQmlComponentPrivate::completeCreate()
 {
-    const RequiredProperties& unsetRequiredProperties = requiredProperties();
-    for (const auto& unsetRequiredProperty: unsetRequiredProperties) {
-        QQmlError error = unsetRequiredPropertyToQQmlError(unsetRequiredProperty);
-        state.errors.push_back(error);
+    if (state.hasUnsetRequiredProperties()) {
+        for (const auto& unsetRequiredProperty: std::as_const(*state.requiredProperties())) {
+            QQmlError error = unsetRequiredPropertyToQQmlError(unsetRequiredProperty);
+            state.errors.push_back(QQmlComponentPrivate::AnnotatedQmlError { error, true });
+        }
     }
-    if (state.completePending) {
+    if (loadedType.isValid()) {
+        /*
+           We can directly set completePending to false, as finalize is only concerned
+           with setting up pending bindings, but that cannot happen here, as we're
+           dealing with a pure C++ type, which cannot have pending bindings
+        */
+        state.setCompletePending(false);
+        QQmlEnginePrivate::get(engine)->inProgressCreations--;
+    } else if (state.isCompletePending()) {
         ++creationDepth;
         QQmlEnginePrivate *ep = QQmlEnginePrivate::get(engine);
         complete(ep, &state);
@@ -1160,6 +1245,80 @@ QQmlComponentAttached *QQmlComponent::qmlAttachedProperties(QObject *obj)
     return a;
 }
 
+/*!
+    Load the QQmlComponent for \a typeName in the module \a uri.
+    If the type is implemented via a QML file, \a mode is used to
+    load it. Types backed by C++ are always loaded synchronously.
+
+    \code
+    QQmlEngine engine;
+    QQmlComponent component(&engine);
+    component.loadFromModule("QtQuick", "Item");
+    // once the component is ready
+    std::unique_ptr<QObject> item(component.create());
+    Q_ASSERT(item->metaObject() == &QQuickItem::staticMetaObject);
+    \endcode
+
+    \since 6.5
+    \sa loadUrl()
+ */
+void QQmlComponent::loadFromModule(QAnyStringView uri, QAnyStringView typeName,
+                                   QQmlComponent::CompilationMode mode)
+{
+    Q_D(QQmlComponent);
+
+    auto enginePriv = QQmlEnginePrivate::get(d->engine);
+    // LoadHelper must be on the Heap as it derives from QQmlRefCount
+    auto loadHelper = QQml::makeRefPointer<LoadHelper>(&enginePriv->typeLoader, uri);
+
+    auto [moduleStatus, type] = loadHelper->resolveType(typeName);
+    auto reportError = [&](QString msg) {
+        QQmlError error;
+        error.setDescription(msg);
+        d->state.errors.push_back(std::move(error));
+        emit statusChanged(Error);
+    };
+    if (moduleStatus == LoadHelper::ResolveTypeResult::NoSuchModule) {
+        reportError(QLatin1String(R"(No module named "%1" found)")
+                    .arg(uri.toString()));
+    } else if (!type.isValid()) {
+        reportError(QLatin1String(R"(Module "%1" contains no type named "%2")")
+                    .arg(uri.toString(), typeName.toString()));
+    } else if (type.isCreatable()) {
+        d->clear();
+        // mimic the progressChanged behavior from loadUrl
+        if (d->progress != 0) {
+            d->progress = 0;
+            emit progressChanged(0);
+        }
+        d->loadedType = type;
+        d->progress = 1;
+        emit progressChanged(1);
+        emit statusChanged(status());
+
+    } else if (type.isComposite()) {
+        loadUrl(type.sourceUrl(), mode);
+    } else if (type.isInlineComponentType()) {
+        auto baseUrl = type.sourceUrl();
+        baseUrl.setFragment(QString());
+        // if the outer type has not been resolved yet, we need to load the outer type synchronously
+        // in order to get the correct object id
+        mode = type.inlineComponentObjectId() > 0 ? mode : QQmlComponent::CompilationMode::PreferSynchronous;
+        loadUrl(baseUrl, mode);
+        if (!isError()) {
+            d->isInlineComponent = true;
+            d->start = type.inlineComponentObjectId();
+            Q_ASSERT(d->start >= 0);
+        }
+    } else if (type.isSingleton() || type.isCompositeSingleton()) {
+        reportError(QLatin1String(R"(%1 is a singleton, and cannot be loaded)")
+                    .arg(typeName.toString()));
+    } else {
+        reportError(QLatin1String("Could not load %1, as the type is uncreatable")
+                                 .arg(typeName.toString()));
+    }
+}
+
 /*!
     Create an object instance from this component using the provided
     \a incubator. \a context specifies the context within which to create the object
@@ -1209,6 +1368,14 @@ void QQmlComponent::create(QQmlIncubator &incubator, QQmlContext *context, QQmlC
     incubator.clear();
     QExplicitlySharedDataPointer<QQmlIncubatorPrivate> p(incubator.d);
 
+    if (d->loadedType.isValid()) {
+        // there isn't really an incubation process for C++ backed types
+        // so just create the object and signal that we are ready
+
+        p->incubateCppBasedComponent(this, context);
+        return;
+    }
+
     QQmlEnginePrivate *enginePriv = QQmlEnginePrivate::get(d->engine);
 
     p->compilationUnit = d->compilationUnit;
@@ -1300,7 +1467,7 @@ struct QmlIncubatorObject : public QV4::Object
     static ReturnedValue method_forceCompletion(const FunctionObject *, const Value *thisObject, const Value *argv, int argc);
 
     void statusChanged(QQmlIncubator::Status);
-    void setInitialState(QObject *, RequiredProperties &requiredProperties);
+    void setInitialState(QObject *, RequiredProperties *requiredProperties);
 };
 
 }
@@ -1351,8 +1518,7 @@ static void QQmlComponent_setQmlParent(QObject *me, QObject *parent)
             }
         }
         if (needParent)
-            qWarning("QQmlComponent: Created graphical object was not "
-                     "placed in the graphics scene.");
+            qmlWarning(me) << "Created graphical object was not placed in the graphics scene.";
     }
 }
 
@@ -1399,7 +1565,7 @@ static void QQmlComponent_setQmlParent(QObject *me, QObject *parent)
 */
 
 
-void QQmlComponentPrivate::setInitialProperties(QV4::ExecutionEngine *engine, QV4::QmlContext *qmlContext, const QV4::Value &o, const QV4::Value &v, RequiredProperties &requiredProperties, QObject *createdComponent)
+void QQmlComponentPrivate::setInitialProperties(QV4::ExecutionEngine *engine, QV4::QmlContext *qmlContext, const QV4::Value &o, const QV4::Value &v, RequiredProperties *requiredProperties, QObject *createdComponent)
 {
     QV4::Scope scope(engine);
     QV4::ScopedObject object(scope);
@@ -1444,7 +1610,7 @@ void QQmlComponentPrivate::setInitialProperties(QV4::ExecutionEngine *engine, QV
         if (engine->hasException) {
             qmlWarning(createdComponent, engine->catchExceptionAsQmlError());
             continue;
-        } else if (isTopLevelProperty) {
+        } else if (isTopLevelProperty && requiredProperties) {
             auto prop = removePropertyFromRequired(createdComponent, name->toQString(),
                                                    requiredProperties, engine->qmlEngine());
         }
@@ -1533,11 +1699,12 @@ void QQmlComponent::createObject(QQmlV4Function *args)
 
     if (!valuemap->isUndefined()) {
         QV4::Scoped<QV4::QmlContext> qmlContext(scope, v4->qmlContext());
-        QQmlComponentPrivate::setInitialProperties(v4, qmlContext, object, valuemap, d->requiredProperties(), rv);
+        QQmlComponentPrivate::setInitialProperties(v4, qmlContext, object, valuemap,
+                                                   d->state.requiredProperties(), rv);
     }
-    if (!d->requiredProperties().empty()) {
+    if (d->state.hasUnsetRequiredProperties()) {
         QList<QQmlError> errors;
-        for (const auto &requiredProperty: d->requiredProperties()) {
+        for (const auto &requiredProperty: std::as_const(*d->state.requiredProperties())) {
             errors.push_back(QQmlComponentPrivate::unsetRequiredPropertyToQQmlError(requiredProperty));
         }
         qmlWarning(rv, errors);
@@ -1563,46 +1730,14 @@ QObject *QQmlComponent::createObject(QObject *parent, const QVariantMap &propert
 {
     Q_D(QQmlComponent);
     Q_ASSERT(d->engine);
-
-    QQmlContext *ctxt = creationContext();
-    if (!ctxt)
-        ctxt = d->engine->rootContext();
-
-    QObject *rv = beginCreate(ctxt);
-    if (!rv)
-        return nullptr;
-
-    Q_ASSERT(d->state.errors.isEmpty()); // otherwise beginCreate() would return nullptr
-
-    QQmlComponent_setQmlParent(rv, parent);
-
-    if (!properties.isEmpty()) {
-        setInitialProperties(rv, properties);
-        if (!d->state.errors.isEmpty()) {
-            qmlWarning(rv, d->state.errors);
-            d->state.errors.clear();
-        }
+    QObject *rv = d->createWithProperties(parent, properties, creationContext(),
+                                          QQmlComponentPrivate::CreateWarnAboutRequiredProperties);
+    if (rv) {
+        QQmlData *qmlData = QQmlData::get(rv);
+        Q_ASSERT(qmlData);
+        qmlData->explicitIndestructibleSet = false;
+        qmlData->indestructible = false;
     }
-
-    if (!d->requiredProperties().empty()) {
-        QList<QQmlError> errors;
-        for (const auto &requiredProperty: std::as_const(d->requiredProperties())) {
-            errors.push_back(QQmlComponentPrivate::unsetRequiredPropertyToQQmlError(
-                    requiredProperty));
-        }
-        if (!errors.isEmpty())
-            qmlWarning(rv, errors);
-        delete rv;
-        return nullptr;
-    }
-
-    d->completeCreate();
-
-    QQmlData *qmlData = QQmlData::get(rv);
-    Q_ASSERT(qmlData);
-    qmlData->explicitIndestructibleSet = false;
-    qmlData->indestructible = false;
-
     return rv;
 }
 
@@ -1726,7 +1861,7 @@ void QQmlComponent::incubateObject(QQmlV4Function *args)
 }
 
 // XXX used by QSGLoader
-void QQmlComponentPrivate::initializeObjectWithInitialProperties(QV4::QmlContext *qmlContext, const QV4::Value &valuemap, QObject *toCreate, RequiredProperties &requiredProperties)
+void QQmlComponentPrivate::initializeObjectWithInitialProperties(QV4::QmlContext *qmlContext, const QV4::Value &valuemap, QObject *toCreate, RequiredProperties *requiredProperties)
 {
     QV4::ExecutionEngine *v4engine = engine->handle();
     QV4::Scope scope(v4engine);
@@ -1825,7 +1960,7 @@ void QV4::Heap::QmlIncubatorObject::destroy() {
     Object::destroy();
 }
 
-void QV4::QmlIncubatorObject::setInitialState(QObject *o, RequiredProperties &requiredProperties)
+void QV4::QmlIncubatorObject::setInitialState(QObject *o, RequiredProperties *requiredProperties)
 {
     QQmlComponent_setQmlParent(o, d()->parent);
 
