@@ -85,12 +85,6 @@ QTypeRevision relevantVersion(const QString &uri, QTypeRevision version)
     return QQmlMetaType::latestModuleVersion(uri).isValid() ? version : QTypeRevision();
 }
 
-QTypeRevision validVersion(QTypeRevision version = QTypeRevision())
-{
-    // If the given version is invalid, return a valid but useless version to signal "It's OK".
-    return version.isValid() ? version : QTypeRevision::fromMinorVersion(0);
-}
-
 QQmlError moduleNotFoundError(const QString &uri, QTypeRevision version)
 {
     QQmlError error;
@@ -203,6 +197,12 @@ bool isPathAbsolute(const QString &path)
 \internal
 */
 
+QTypeRevision QQmlImports::validVersion(QTypeRevision version)
+{
+    // If the given version is invalid, return a valid but useless version to signal "It's OK".
+    return version.isValid() ? version : QTypeRevision::fromMinorVersion(0);
+}
+
 /*!
   Sets the base URL to be used for all relative file imports added.
 */
@@ -1221,7 +1221,7 @@ QQmlImportNamespace *QQmlImports::importNamespace(const QString &prefix)
 
 QQmlImportInstance *QQmlImports::addImportToNamespace(
         QQmlImportNamespace *nameSpace, const QString &uri, const QString &url, QTypeRevision version,
-        QV4::CompiledData::Import::ImportType type, QList<QQmlError> *errors, ImportFlags flags)
+        QV4::CompiledData::Import::ImportType type, QList<QQmlError> *errors, quint16 precedence)
 {
     Q_ASSERT(nameSpace);
     Q_ASSERT(errors);
@@ -1233,38 +1233,25 @@ QQmlImportInstance *QQmlImports::addImportToNamespace(
     import->url = url;
     import->version = version;
     import->isLibrary = (type == QV4::CompiledData::Import::ImportLibrary);
+    import->precedence = precedence;
+    import->implicitlyImported = precedence >= QQmlImportInstance::Implicit;
 
-    if (flags & QQmlImports::ImportImplicit) {
-        import->implicitlyImported = true;
-        nameSpace->imports.append(import);
+    for (auto it = nameSpace->imports.cbegin(), end = nameSpace->imports.cend();
+         it != end; ++it) {
+        if ((*it)->precedence < precedence)
+            continue;
+
+        nameSpace->imports.insert(it, import);
         return import;
     }
-
-    if (flags & QQmlImports::ImportLowPrecedence) {
-        for (auto it = nameSpace->imports.rbegin(), end = nameSpace->imports.rend();
-             it != end; ++it) {
-            if (!(*it)->implicitlyImported) {
-                nameSpace->imports.insert(it.base(), import);
-                return import;
-            }
-        }
-    }
-
-    // This is one of 3 cases:
-    //
-    // 1. existing imports are empty
-    // 2. new import is low precedence and all existing ones are implicit
-    // 3. new import is normal precedence
-    //
-    // In those cases the new import overrides all existing ones and has to be prepended.
-    nameSpace->imports.prepend(import);
+    nameSpace->imports.append(import);
     return import;
 }
 
 QTypeRevision QQmlImports::addLibraryImport(
         QQmlImportDatabase *database, const QString &uri, const QString &prefix,
         QTypeRevision version, const QString &qmldirIdentifier, const QString &qmldirUrl,
-        ImportFlags flags, QList<QQmlError> *errors)
+        ImportFlags flags, quint16 precedence, QList<QQmlError> *errors)
 {
     Q_ASSERT(database);
     Q_ASSERT(errors);
@@ -1279,7 +1266,7 @@ QTypeRevision QQmlImports::addLibraryImport(
     QQmlImportInstance *inserted = addImportToNamespace(
                 nameSpace, uri, qmldirUrl, version,
                 QV4::CompiledData::Import::ImportLibrary, errors,
-                flags);
+                precedence);
     Q_ASSERT(inserted);
 
     if (!(flags & QQmlImports::ImportIncomplete)) {
@@ -1347,7 +1334,8 @@ QTypeRevision QQmlImports::addLibraryImport(
 */
 QTypeRevision QQmlImports::addFileImport(
         QQmlImportDatabase *database, const QString &uri, const QString &prefix,
-        QTypeRevision version, ImportFlags flags, QString *localQmldir, QList<QQmlError> *errors)
+        QTypeRevision version, ImportFlags flags, quint16 precedence,
+        QString *localQmldir, QList<QQmlError> *errors)
 {
     Q_ASSERT(database);
     Q_ASSERT(errors);
@@ -1390,7 +1378,7 @@ QTypeRevision QQmlImports::addFileImport(
 
         const QString dir = localFileOrQrc.left(localFileOrQrc.lastIndexOf(Slash) + 1);
         if (!m_typeLoader->directoryExists(dir)) {
-            if (!(flags & QQmlImports::ImportImplicit)) {
+            if (precedence < QQmlImportInstance::Implicit) {
                 QQmlError error;
                 error.setDescription(QQmlImportDatabase::tr("\"%1\": no such directory").arg(uri));
                 error.setUrl(QUrl(qmldirUrl));
@@ -1413,7 +1401,7 @@ QTypeRevision QQmlImports::addFileImport(
 
     } else if (nameSpace->prefix.isEmpty() && !(flags & QQmlImports::ImportIncomplete)) {
 
-        if (!(flags & QQmlImports::ImportImplicit)) {
+        if (precedence < QQmlImportInstance::Implicit) {
             QQmlError error;
             error.setDescription(QQmlImportDatabase::tr("import \"%1\" has no qmldir and no namespace").arg(importUri));
             error.setUrl(QUrl(qmldirUrl));
@@ -1433,7 +1421,7 @@ QTypeRevision QQmlImports::addFileImport(
     //     if the implicit import has already been explicitly added, otherwise we can run into issues
     //     with duplicate imports. However remember that we attempted to add this as implicit import, to
     //     allow for the loading of internal types.
-    if (flags & QQmlImports::ImportImplicit) {
+    if (precedence >= QQmlImportInstance::Implicit) {
         for (QList<QQmlImportInstance *>::const_iterator it = nameSpace->imports.constBegin();
              it != nameSpace->imports.constEnd(); ++it) {
             if ((*it)->url == url) {
@@ -1443,25 +1431,21 @@ QTypeRevision QQmlImports::addFileImport(
         }
     }
 
-    QQmlImportInstance *inserted = addImportToNamespace(
-                nameSpace, importUri, url, version, QV4::CompiledData::Import::ImportFile,
-                errors, flags);
-    Q_ASSERT(inserted);
-    if (flags & QQmlImports::ImportImplicit)
-        inserted->implicitlyImported = true;
-
     if (!(flags & QQmlImports::ImportIncomplete) && !qmldirIdentifier.isEmpty()) {
         QQmlTypeLoaderQmldirContent qmldir;
         if (!getQmldirContent(qmldirIdentifier, importUri, &qmldir, errors))
             return QTypeRevision();
 
         if (qmldir.hasContent()) {
-            if (uri == QStringLiteral(".")) {
-                // If this is an implicit import, prefer the qmldir URI. Unless it doesn't exist.
-                const QString qmldirUri = qmldir.typeNamespace();
-                if (!qmldirUri.isEmpty())
-                    importUri = qmldirUri;
-            }
+            // Prefer the qmldir URI. Unless it doesn't exist.
+            const QString qmldirUri = qmldir.typeNamespace();
+            if (!qmldirUri.isEmpty())
+                importUri = qmldirUri;
+
+            QQmlImportInstance *inserted = addImportToNamespace(
+                        nameSpace, importUri, url, version, QV4::CompiledData::Import::ImportFile,
+                        errors, precedence);
+            Q_ASSERT(inserted);
 
             version = importExtension(importUri, version, database, &qmldir, errors);
             if (!version.isValid())
@@ -1469,9 +1453,15 @@ QTypeRevision QQmlImports::addFileImport(
 
             if (!inserted->setQmldirContent(url, qmldir, nameSpace, errors))
                 return QTypeRevision();
+
+            return validVersion(version);
         }
     }
 
+    QQmlImportInstance *inserted = addImportToNamespace(
+                nameSpace, importUri, url, version, QV4::CompiledData::Import::ImportFile,
+                errors, precedence);
+    Q_ASSERT(inserted);
     return validVersion(version);
 }
 
@@ -1601,7 +1591,14 @@ QQmlImportDatabase::QQmlImportDatabase(QQmlEngine *e)
 : engine(e)
 {
     filePluginPath << QLatin1String(".");
-    // Search order is applicationDirPath(), qrc:/qt-project.org/imports, $QML_IMPORT_PATH, $QML2_IMPORT_PATH, QLibraryInfo::QmlImportsPath
+    // Search order is:
+    // 1. android or macos specific bundle paths.
+    // 2. applicationDirPath()
+    // 3. qrc:/qt-project.org/imports
+    // 4. qrc:/qt/qml
+    // 5. $QML2_IMPORT_PATH
+    // 6. $QML_IMPORT_PATH
+    // 7. QLibraryInfo::QmlImportsPath
 
     QString installImportsPath = QLibraryInfo::path(QLibraryInfo::QmlImportsPath);
     addImportPath(installImportsPath);
@@ -1618,6 +1615,7 @@ QQmlImportDatabase::QQmlImportDatabase(QQmlEngine *e)
     addEnvImportPath("QML_IMPORT_PATH");
     addEnvImportPath("QML2_IMPORT_PATH");
 
+    addImportPath(QStringLiteral("qrc:/qt/qml"));
     addImportPath(QStringLiteral("qrc:/qt-project.org/imports"));
     addImportPath(QCoreApplication::applicationDirPath());
 
